/**
  Raytracing demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2008-02-28 (Public Domain)
*/
#include <GL/glew.h> /*<- for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <iostream>
#include <GL/glut.h>
#include <math.h>
#include "osl/vec4.h" /*<- handy little vector class */
/* Convert a struct with .x, .y, .z fields into three arguments */
#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z


#include "image.h" /* (originally from) Nigel Stewart's GLT image library */
#include "image.cpp"

#include "glsl.h" /* Orion's tiny GLSL library */
#include "glsl.cpp"

/*********************** App Code **********************/

/* Orthonormal coordinate frame */
class ortho_frame {
public:
	/* Unit vectors pointing along axes of our frame.
		X cross Y is Z. (right-handed coordinate system)
	*/
	vec3 x,y,z;
	ortho_frame() :x(1,0,0), y(0,1,0), z(0,0,1) {}
	
	/* Reorient this coordinate frame by this far 
	   along in the X and Y axes.  "dx" is the distance
	   along the z axis to push the x axis; "dy" 
	   the distance along the z axis to push the y axis.
	*/
	void nudge(double dx,double dy) {
		x+=dx*z;
		y+=dy*z;
		orthonormalize();
	}
	
	/* Reconstruct orthonormal frame from X and Y axes.
	  Y is primary, X is secondary, Z is tertiary.
	*/
	void orthonormalize(void) {
		y=normalize(y);
		z=normalize(cross(x,y));
		x=normalize(cross(y,z));
	}
};

/* Orientation of our camera. Camera looks down -Z axis; Y is up. */
ortho_frame camera_frame;

void my_display(void) {
	glMatrixMode(GL_PROJECTION); /* set up projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT), /* aspect ratio: width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	vec3 camera_location=1.3*camera_frame.z; /* back away from origin */
	gluLookAt(VEC3_TO_XYZ(camera_location),
		VEC3_TO_XYZ(vec3(0)),
		VEC3_TO_XYZ(camera_frame.y));
	
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	glClearColor(0.3,0.5,0.7,0); /* set background color to blue */
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	
	static GLhandleARB prog=
		makeProgramObjectFromFiles("vertex.txt","fragment.txt");
	
	glUseProgramObjectARB(prog);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	glUniform3fvARB( /* set the "cameraCoords" uniform variable */
		glGetUniformLocationARB(prog, "cameraLoc"),
		1,
		camera_location
	);
	
	/* Fill the screen with raytraced pixels.
	   The specific geometry you use here doesn't matter--
	     the displayed image will totally come from the raytracer! */
	glutSolidSphere(1.9, 5,3); 
	
	glUseProgramObjectARB(0); /* back to fixed-function */
	
	glutSwapBuffers(); /* copy back buffer to front */

	
/* Statistics */
	static int last_msec=0;
	int cur_msec=glutGet(GLUT_ELAPSED_TIME); /* milliseconds since glutInit */
	int msec_elapsed=cur_msec-last_msec;
	printf("%d msec/frame; %.2f fps\n",
		msec_elapsed,1.0e3/msec_elapsed);
	last_msec=cur_msec;
}

int mouse_x, mouse_y;
void my_mouse(int state,int buttons,int x,int y) {
	mouse_x=x; mouse_y=y;  /* just stash motion */
}
void my_motion(int x,int y) { 
	float dx=x-mouse_x, dy=y-mouse_y; /* relative motion */
	
	float speed=1.0/100; /* radians rotation per pixel */
	camera_frame.nudge(speed*dx,-speed*dy); /* OpenGL Y is negative of mouse Y! */
	
	mouse_x=x; mouse_y=y; 
	glutPostRedisplay();
}

int main(int argc,char *argv[]) {
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
        glutInitDisplayMode (mode);
	glutCreateWindow("GLSL Raytracing Demo");
	glewInit();
	
	glutDisplayFunc(my_display);
	glutMotionFunc(my_motion);
	glutMouseFunc(my_mouse);
	
	glutMainLoop(); /* processes events */
	return 0;
}
